@@ -1,7 +1,8 @@
 /* Configuration file handling.
  *
- * Services is copyright (c) 1996-1999 Andy Church.
- *     E-mail: <achurch@dragonfire.net>
+ * IRC Services is copyright (c) 1996-2001 Andrew Church.
+ *     E-mail: <achurch@achurch.org>
+ * Parts copyright (c) 1999-2000 Andrew Kempe and others.
  * This program is free but copyrighted software; see the file COPYING for
  * details.
  */
@@ -22,6 +23,7 @@ char *ServerName;
 char *ServerDesc;
 char *ServiceUser;
 char *ServiceHost;
+int   ServerNumeric;
 static char *temp_userhost;
 
 char *s_NickServ;
@@ -29,6 +31,7 @@ char *s_ChanServ;
 char *s_MemoServ;
 char *s_HelpServ;
 char *s_OperServ;
+char *s_StatServ;
 char *s_GlobalNoticer;
 char *s_IrcIIHelp;
 char *s_DevNull;
@@ -37,6 +40,7 @@ char *desc_ChanServ;
 char *desc_MemoServ;
 char *desc_HelpServ;
 char *desc_OperServ;
+char *desc_StatServ;
 char *desc_GlobalNoticer;
 char *desc_IrcIIHelp;
 char *desc_DevNull;
@@ -47,10 +51,12 @@ char *HelpDir;
 char *NickDBName;
 char *ChanDBName;
 char *OperDBName;
+char *StatDBName;
 char *AutokillDBName;
 char *NewsDBName;
 
 int   NoBackupOkay;
+int   NoBouncyModes;
 int   NoSplitRecovery;
 int   StrictPasswords;
 int   BadPassLimit;
@@ -60,31 +66,26 @@ int   ExpireTimeout;
 int   ReadTimeout;
 int   WarningTimeout;
 int   TimeoutCheck;
+int   PingFrequency;
+int   MergeChannelModes;
 
-static int NSDefNone;
 int   NSForceNickChange;
 char *NSGuestNickPrefix;
-int   NSDefKill;
-int   NSDefKillQuick;
-int   NSDefSecure;
-int   NSDefPrivate;
-int   NSDefHideEmail;
-int   NSDefHideUsermask;
-int   NSDefHideQuit;
-int   NSDefMemoSignon;
-int   NSDefMemoReceive;
+int   NSDefFlags;
 int   NSRegDelay;
+int   NSRequireEmail;
 int   NSExpire;
 int   NSAccessMax;
 char *NSEnforcerUser;
 char *NSEnforcerHost;
-static char *temp_nsuserhost;
 int   NSReleaseTimeout;
 int   NSAllowKillImmed;
-int   NSDisableLinkCommand;
+int   NSMaxLinkDepth;
 int   NSListOpersOnly;
 int   NSListMax;
 int   NSSecureAdmins;
+int   NSSuspendExpire;
+int   NSSuspendGrace;
 
 int   CSMaxReg;
 int   CSExpire;
@@ -95,6 +96,8 @@ int   CSInhabit;
 int   CSRestrictDelay;
 int   CSListOpersOnly;
 int   CSListMax;
+int   CSSuspendExpire;
+int   CSSuspendGrace;
 
 int   MSMaxMemos;
 int   MSSendDelay;
@@ -102,13 +105,15 @@ int   MSNotifyAll;
 
 char *ServicesRoot;
 int   LogMaxUsers;
+char *StaticAkillReason;
+int   ImmediatelySendAkill;
 int   AutokillExpiry;
 int   WallOper;
 int   WallBadOS;
-int   WallOSMode;
-int   WallOSClearmodes;
-int   WallOSKick;
+int   WallOSChannel;
 int   WallOSAkill;
+int   WallOSException;
+int   WallSU;
 int   WallAkillExpire;
 int   WallExceptionExpire;
 int   WallGetpass;
@@ -121,6 +126,8 @@ int   KillClones;
 
 int   KillClonesAkillExpire;
 
+int   SSOpersOnly;
+
 int   LimitSessions;
 int   DefSessionLimit;
 int   ExceptionExpiry;
@@ -128,17 +135,31 @@ int   MaxSessionLimit;
 char *ExceptionDBName;
 char *SessionLimitExceeded;
 char *SessionLimitDetailsLoc;
+int   SessionLimitAkill;
+int   SessionLimitMinKillTime;
+int   SessionLimitMaxKillCount;
+int   SessionLimitAkillExpiry;
+char *SessionLimitAkillReason;
+
+/******* Local use only: *******/
+
+static int   NSDefNone;
+static int   NSDefKill;
+static int   NSDefKillQuick;
+static int   NSDefSecure;
+static int   NSDefPrivate;
+static int   NSDefHideEmail;
+static int   NSDefHideUsermask;
+static int   NSDefHideQuit;
+static int   NSDefMemoSignon;
+static int   NSDefMemoReceive;
+static int   NSDisableLinkCommand;
+static char *temp_nsuserhost;
 
 /*************************************************************************/
 
 /* Deprecated directive (dep_) and value checking (chk_) functions: */
 
-static void dep_ListOpersOnly(void)
-{
-    NSListOpersOnly = 1;
-    CSListOpersOnly = 1;
-}
-
 /*************************************************************************/
 
 #define MAXPARAMS	8
@@ -160,6 +181,8 @@ typedef struct {
 #define PARAM_PORT	3	/* 1..65535 only */
 #define PARAM_STRING	4
 #define PARAM_TIME	5
+#define PARAM_TIMEMSEC	6	/* Variable is in milliseconds, parameter
+				 *    in seconds (decimal allowed) */
 #define PARAM_SET	-1	/* Not a real parameter; just set the
 				 *    given integer variable to 1 */
 #define PARAM_DEPRECATED -2	/* Set for deprecated directives; `ptr'
@@ -177,7 +200,8 @@ Directive directives[] = {
     { "ChanServDB",       { { PARAM_STRING, 0, &ChanDBName } } },
     { "ChanServName",     { { PARAM_STRING, 0, &s_ChanServ },
                             { PARAM_STRING, 0, &desc_ChanServ } } },
-    { "CheckClones",      { { PARAM_SET, PARAM_FULLONLY, &CheckClones },
+    { "CheckClones",      { { PARAM_DEPRECATED, PARAM_FULLONLY, NULL },
+			    { PARAM_SET, 0, &CheckClones },
                             { PARAM_POSINT, 0, &CloneMinUsers },
                             { PARAM_TIME, 0, &CloneMaxDelay },
                             { PARAM_TIME, 0, &CloneWarningDelay } } },
@@ -190,7 +214,9 @@ Directive directives[] = {
     { "CSListOpersOnly",  { { PARAM_SET, 0, &CSListOpersOnly } } },
     { "CSMaxReg",         { { PARAM_POSINT, 0, &CSMaxReg } } },
     { "CSRestrictDelay",  { { PARAM_TIME, 0, &CSRestrictDelay } } },
-    { "DefSessionLimit",  { { PARAM_POSINT, 0, &DefSessionLimit } } },
+    { "CSSuspendExpire",  { { PARAM_TIME, 0 , &CSSuspendExpire },
+                            { PARAM_TIME, 0 , &CSSuspendGrace } } },
+    { "DefSessionLimit",  { { PARAM_INT, 0, &DefSessionLimit } } },
     { "DevNullName",      { { PARAM_STRING, 0, &s_DevNull },
                             { PARAM_STRING, 0, &desc_DevNull } } },
     { "ExceptionDB",	  { { PARAM_STRING, 0, &ExceptionDBName } } },
@@ -201,18 +227,20 @@ Directive directives[] = {
     { "HelpDir",          { { PARAM_STRING, 0, &HelpDir } } },
     { "HelpServName",     { { PARAM_STRING, 0, &s_HelpServ },
                             { PARAM_STRING, 0, &desc_HelpServ } } },
+    { "ImmediatelySendAkill",{{PARAM_SET, 0, &ImmediatelySendAkill } } },
     { "IrcIIHelpName",    { { PARAM_STRING, 0, &s_IrcIIHelp },
                             { PARAM_STRING, 0, &desc_IrcIIHelp } } },
-    { "KillClones",       { { PARAM_SET, PARAM_FULLONLY, &KillClones } } },
+    { "KillClones",       { { PARAM_DEPRECATED, PARAM_FULLONLY, NULL },
+			    { PARAM_SET, 0, &KillClones } } },
     { "KillClonesAkillExpire",{{PARAM_TIME, 0, &KillClonesAkillExpire } } },
     { "LimitSessions",    { { PARAM_SET, PARAM_FULLONLY, &LimitSessions } } },
-    { "ListOpersOnly",    { { PARAM_DEPRECATED, 0, dep_ListOpersOnly } } },
     { "LocalAddress",     { { PARAM_STRING, 0, &LocalHost },
                             { PARAM_PORT, PARAM_OPTIONAL, &LocalPort } } },
     { "LogMaxUsers",      { { PARAM_SET, 0, &LogMaxUsers } } },
     { "MaxSessionLimit",  { { PARAM_POSINT, 0, &MaxSessionLimit } } },
     { "MemoServName",     { { PARAM_STRING, 0, &s_MemoServ },
                             { PARAM_STRING, 0, &desc_MemoServ } } },
+    { "MergeChannelModes",{ { PARAM_TIMEMSEC, 0, &MergeChannelModes } } },
     { "MOTDFile",         { { PARAM_STRING, 0, &MOTDFilename } } },
     { "MSMaxMemos",       { { PARAM_POSINT, 0, &MSMaxMemos } } },
     { "MSNotifyAll",      { { PARAM_SET, 0, &MSNotifyAll } } },
@@ -222,6 +250,7 @@ Directive directives[] = {
     { "NickServName",     { { PARAM_STRING, 0, &s_NickServ },
                             { PARAM_STRING, 0, &desc_NickServ } } },
     { "NoBackupOkay",     { { PARAM_SET, 0, &NoBackupOkay } } },
+    { "NoBouncyModes",    { { PARAM_SET, 0, &NoBouncyModes } } },
     { "NoSplitRecovery",  { { PARAM_SET, 0, &NoSplitRecovery } } },
     { "NSAccessMax",      { { PARAM_POSINT, 0, &NSAccessMax } } },
     { "NSAllowKillImmed", { { PARAM_SET, 0, &NSAllowKillImmed } } },
@@ -242,25 +271,41 @@ Directive directives[] = {
     { "NSGuestNickPrefix",{ { PARAM_STRING, 0, &NSGuestNickPrefix } } },
     { "NSListMax",        { { PARAM_POSINT, 0, &NSListMax } } },
     { "NSListOpersOnly",  { { PARAM_SET, 0, &NSListOpersOnly } } },
+    { "NSMaxLinkDepth",   { { PARAM_INT, 0, &NSMaxLinkDepth } } },
     { "NSRegDelay",       { { PARAM_TIME, 0, &NSRegDelay } } },
+    { "NSRequireEmail",   { { PARAM_SET, 0, &NSRequireEmail } } },
     { "NSReleaseTimeout", { { PARAM_TIME, 0, &NSReleaseTimeout } } },
     { "NSSecureAdmins",   { { PARAM_SET, 0, &NSSecureAdmins } } },
+    { "NSSuspendExpire",  { { PARAM_TIME, 0 , &NSSuspendExpire },
+                            { PARAM_TIME, 0 , &NSSuspendGrace } } },
     { "OperServDB",       { { PARAM_STRING, 0, &OperDBName } } },
     { "OperServName",     { { PARAM_STRING, 0, &s_OperServ },
                             { PARAM_STRING, 0, &desc_OperServ } } },
     { "PIDFile",          { { PARAM_STRING, 0, &PIDFilename } } },
+    { "PingFrequency",    { { PARAM_TIME, 0, &PingFrequency } } },
     { "ReadTimeout",      { { PARAM_TIME, 0, &ReadTimeout } } },
     { "RemoteServer",     { { PARAM_STRING, 0, &RemoteServer },
                             { PARAM_PORT, 0, &RemotePort },
                             { PARAM_STRING, 0, &RemotePassword } } },
     { "ServerDesc",       { { PARAM_STRING, 0, &ServerDesc } } },
     { "ServerName",       { { PARAM_STRING, 0, &ServerName } } },
+    { "ServerNumeric",	  { { PARAM_POSINT, 0, &ServerNumeric } } },
     { "ServicesRoot",     { { PARAM_STRING, 0, &ServicesRoot } } },
     { "ServiceUser",      { { PARAM_STRING, 0, &temp_userhost } } },
-    { "SessionLimitDetailsLoc",{{PARAM_STRING, 0, &SessionLimitDetailsLoc}}},
-    { "SessionLimitExceeded",{{PARAM_STRING, 0, &SessionLimitExceeded}}},
+    { "SessionLimitAkill",{ { PARAM_SET, 0, &SessionLimitAkill },
+    			    { PARAM_TIME, 0, &SessionLimitMinKillTime },
+    			    { PARAM_POSINT, 0, &SessionLimitMaxKillCount },
+    			    { PARAM_TIME, 0, &SessionLimitAkillExpiry } } },
+    { "SessionLimitAkillReason",{{PARAM_STRING, 0, &SessionLimitAkillReason }}},
+    { "SessionLimitDetailsLoc",{{PARAM_STRING, 0, &SessionLimitDetailsLoc } } },
+    { "SessionLimitExceeded",{{PARAM_STRING, 0, &SessionLimitExceeded } } },
+    { "SSOpersOnly",	  { { PARAM_SET, 0, &SSOpersOnly } } },
+    { "StaticAkillReason",{ { PARAM_STRING, 0, &StaticAkillReason } } },
+    { "StatServDB",       { { PARAM_STRING, PARAM_FULLONLY, &StatDBName } } },
+    { "StatServName",     { { PARAM_STRING, PARAM_FULLONLY, &s_StatServ },
+			    { PARAM_STRING, 0, &desc_StatServ } } },
     { "StrictPasswords",  { { PARAM_SET, 0, &StrictPasswords } } },
-    { "TimeoutCheck",     { { PARAM_TIME, 0, &TimeoutCheck } } },
+    { "TimeoutCheck",     { { PARAM_TIMEMSEC, 0, &TimeoutCheck } } },
     { "UpdateTimeout",    { { PARAM_TIME, 0, &UpdateTimeout } } },
     { "WallAkillExpire",  { { PARAM_SET, 0, &WallAkillExpire } } },
     { "WallExceptionExpire",{{PARAM_SET, 0, &WallExceptionExpire } } },
@@ -268,10 +313,10 @@ Directive directives[] = {
     { "WallGetpass",      { { PARAM_SET, 0, &WallGetpass } } },
     { "WallOper",         { { PARAM_SET, 0, &WallOper } } },
     { "WallOSAkill",      { { PARAM_SET, 0, &WallOSAkill } } },
-    { "WallOSClearmodes", { { PARAM_SET, 0, &WallOSClearmodes } } },
-    { "WallOSKick",       { { PARAM_SET, 0, &WallOSKick } } },
-    { "WallOSMode",       { { PARAM_SET, 0, &WallOSMode } } },
+    { "WallOSChannel",    { { PARAM_SET, 0, &WallOSChannel } } },
+    { "WallOSException",  { { PARAM_SET, 0, &WallOSException } } },
     { "WallSetpass",      { { PARAM_SET, 0, &WallSetpass } } },
+    { "WallSU",           { { PARAM_SET, 0, &WallSU } } },
     { "WarningTimeout",   { { PARAM_TIME, 0, &WarningTimeout } } },
 };
 
@@ -368,15 +413,16 @@ int parse(char *buf, int linenum)
 #ifdef STREAMLINED
 	    if (d->params[i].flags & PARAM_FULLONLY) {
 		error(linenum, "Directive `%s' not available in STREAMLINED mode",
-			d->name);
+		      d->name);
 		break;
 	    }
 #endif
 	    if (d->params[i].type == PARAM_DEPRECATED) {
-		void (*func)(void);
+		void (*func)(void); /* For clarity */
 		error(linenum, "Deprecated directive `%s' used", d->name);
 		func = (void (*)(void))(d->params[i].ptr);
-		func();  /* For clarity */
+		if (func)
+		    func();
 		continue;
 	    }
 	    if (optind >= ac) {
@@ -443,6 +489,32 @@ int parse(char *buf, int linenum)
 		}
 		*(int *)d->params[i].ptr = val;
 		break;
+	      case PARAM_TIMEMSEC:
+		val = strtol(av[optind++], &s, 10);
+		if (val < 0) {
+		    error(linenum,
+			"%s: Expected a positive value for parameter %d",
+			d->name, optind);
+		    retval = 0;
+		    break;
+		} else if (val > 1000000) {
+		    error(linenum, "%s: Value too large (maximum 1000000)",
+			  d->name);
+		}
+		val *= 1000;
+		if (*s++ == '.') {
+		    int decimal = 0;
+		    int count = 0;
+		    while (count < 3 && isdigit(*s)) {
+			decimal = decimal*10 + (*s++ - '0');
+			count++;
+		    }
+		    while (count++ < 3)
+			decimal *= 10;
+		    val += decimal;
+		}
+		*(uint32 *)d->params[i].ptr = val;
+		break;
 	      default:
 		error(linenum, "%s: Unknown type %d for param %d",
 				d->name, d->params[i].type, i+1);
@@ -514,6 +586,9 @@ int read_config()
     CHEK2(s_ChanServ, ChanServName);
     CHEK2(s_MemoServ, MemoServName);
     CHEK2(s_HelpServ, HelpServName);
+#ifdef STATISTICS
+    CHEK2(s_StatServ, StatServName);
+#endif
     CHEK2(s_OperServ, OperServName);
     CHEK2(s_GlobalNoticer, GlobalName);
     CHEK2(PIDFilename, PIDFile);
@@ -522,6 +597,9 @@ int read_config()
     CHEK2(NickDBName, NickServDB);
     CHEK2(ChanDBName, ChanServDB);
     CHEK2(OperDBName, OperServDB);
+#ifdef STATISTICS
+    CHEK2(StatDBName, StatServDB);
+#endif
     CHEK2(AutokillDBName, AutokillDB);
     CHEK2(NewsDBName, NewsDB);
     CHEK2(ExceptionDBName, ExceptionDB);
@@ -533,6 +611,7 @@ int read_config()
     CHECK(NSAccessMax);
     CHEK2(temp_nsuserhost, NSEnforcerUser);
     CHECK(NSReleaseTimeout);
+    CHECK(NSMaxLinkDepth);
     CHECK(NSListMax);
     CHECK(CSAccessMax);
     CHECK(CSAutokickMax);
@@ -540,7 +619,49 @@ int read_config()
     CHECK(CSInhabit);
     CHECK(CSListMax);
     CHECK(ServicesRoot);
-    CHECK(AutokillExpiry);
+    if (SessionLimitAkill) {
+	CHECK(SessionLimitMinKillTime);
+	CHECK(SessionLimitAkillExpiry);
+	CHECK(SessionLimitAkillReason);
+    }
+
+#ifdef IRC_UNREAL
+    if (ServerNumeric < 0 || ServerNumeric > 254) {
+    	error(0, "ServerNumeric must be in the range 1..254");
+	retval = 0;
+    }
+#endif
+
+    /* Note: setting expire timeouts at less than 1 day may result in
+     * bogus help messages (see NICK_HELP and CHAN_HELP). */
+    if (NSExpire > 0 && NSExpire < 86400) {
+	error(0, "NSExpire may not be set less than 1 day");
+	retval = 0;
+    }
+    if (CSExpire > 0 && CSExpire < 86400) {
+	error(0, "CSExpire may not be set less than 1 day");
+	retval = 0;
+    }
+
+    if (NSDisableLinkCommand)
+	NSMaxLinkDepth = 0;
+    if (NSMaxLinkDepth > LINK_HARDMAX) {
+	printf("WARNING: NSMaxLinkDepth (%d) > LINK_HARDMAX (%d), resetting to %d\n",
+	       NSMaxLinkDepth, LINK_HARDMAX, LINK_HARDMAX);
+	NSMaxLinkDepth = LINK_HARDMAX;
+    }
+
+    if (LimitSessions && CheckClones) {
+	printf("\
+WARNING: You have enabled both session limiting (config option: LimitSessions)
+and clone detection (config option: CheckClones). These two features do not
+function correctly when running together. Session limiting is preferred.
+");
+#ifndef NOT_MAIN
+	log("*** WARNING: Both LimitSessions and CheckClones are enabled "
+	    "- this is bad! Check your config.");
+#endif
+    }
 
     if (temp_userhost) {
 	if (!(s = strchr(temp_userhost, '@'))) {
@@ -579,22 +700,28 @@ int read_config()
 	NSDefMemoReceive = 1;
     }
 
-    if (LimitSessions) {
-	CHECK(DefSessionLimit);
-	CHECK(MaxSessionLimit);
-	CHECK(ExceptionExpiry);
+    NSDefFlags = 0;
+    if (NSDefKill)
+	NSDefFlags |= NI_KILLPROTECT;
+    if (NSDefKillQuick)
+	NSDefFlags |= NI_KILL_QUICK;
+    if (NSDefSecure)
+	NSDefFlags |= NI_SECURE;
+    if (NSDefPrivate)
+	NSDefFlags |= NI_PRIVATE;
+    if (NSDefHideEmail)
+	NSDefFlags |= NI_HIDE_EMAIL;
+    if (NSDefHideUsermask)
+	NSDefFlags |= NI_HIDE_MASK;
+    if (NSDefHideQuit)
+	NSDefFlags |= NI_HIDE_QUIT;
+    if (NSDefMemoSignon)
+	NSDefFlags |= NI_MEMO_SIGNON;
+    if (NSDefMemoReceive)
+	NSDefFlags |= NI_MEMO_RECEIVE;
 
-	if (CheckClones) {
-	    printf("Warning: You have enabled both session limiting (config "
-	    	"option: LimitSessions)\nand clone detection (config option: "
-		"CheckClones). These two features do not\nfunction correctly "
-		"when running together. Session limiting is preferred.\n\n");
-#ifndef NOT_MAIN
-	    log("*** Warning: Both LimitSessions and CheckClones are enabled "
-	    		"- this is bad! Check your config.");
-#endif
-	}
-    }
+    if (!MaxSessionLimit)
+	MaxSessionLimit = 32767;
 
     return retval;
 }
