@@ -1,7 +1,8 @@
 /* MemoServ functions.
  *
- * Services is copyright (c) 1996-1999 Andy Church.
- *     E-mail: <achurch@dragonfire.net>
+ * IRC Services is copyright (c) 1996-2001 Andrew Church.
+ *     E-mail: <achurch@achurch.org>
+ * Parts copyright (c) 1999-2000 Andrew Kempe and others.
  * This program is free but copyrighted software; see the file COPYING for
  * details.
  */
@@ -11,7 +12,21 @@
 
 /*************************************************************************/
 
-static int delmemo(MemoInfo *mi, int num);
+/* Error codes for getmemoinfo(). */
+#define GMI_NOTFOUND	-1
+#define GMI_FORBIDDEN	-2
+#define GMI_SUSPENDED	-3
+
+/*************************************************************************/
+
+static MemoInfo *getmemoinfo(const char *name, int *ischan, int *isverboten);
+static int list_memo(User *u, int index, MemoInfo *mi, int *sent_header,
+			int new, const char *chan);
+static int list_memo_callback(User *u, int num, va_list args);
+static int read_memo(User *u, int index, MemoInfo *mi, const char *chan);
+static int read_memo_callback(User *u, int num, va_list args);
+static int del_memo(MemoInfo *mi, int num);
+static int del_memo_callback(User *u, int num, va_list args);
 
 static void do_help(User *u);
 static void do_send(User *u);
@@ -33,7 +48,9 @@ static Command cmds[] = {
     { "DEL",        do_del,  NULL,  MEMO_HELP_DEL,           -1,-1,-1,-1 },
     { "SET",        do_set,  NULL,  MEMO_HELP_SET,           -1,-1,-1,-1 },
     { "SET NOTIFY", NULL,    NULL,  MEMO_HELP_SET_NOTIFY,    -1,-1,-1,-1 },
-    { "SET LIMIT",  NULL,    NULL,  MEMO_HELP_SET_LIMIT,     -1,-1,-1,-1 },
+    { "SET LIMIT",  NULL,    NULL,  -1,
+    		MEMO_HELP_SET_LIMIT, MEMO_SERVADMIN_HELP_SET_LIMIT,
+		MEMO_SERVADMIN_HELP_SET_LIMIT, MEMO_SERVADMIN_HELP_SET_LIMIT },
     { "INFO",       do_info, NULL,  -1,
 		MEMO_HELP_INFO, MEMO_SERVADMIN_HELP_INFO,
 		MEMO_SERVADMIN_HELP_INFO, MEMO_SERVADMIN_HELP_INFO },
@@ -93,6 +110,36 @@ void memoserv(const char *source, char *buf)
 
 /*************************************************************************/
 
+/* Return memo stats.  Assumes pointers are valid. */
+
+void get_memoserv_stats(long *nrec, long *memuse)
+{
+    long count = 0, mem = 0;
+    int i;
+    NickInfo *ni;
+    ChannelInfo *ci;
+
+    for (ni = firstnick(); ni; ni = nextnick()) {
+	mem += ni->memos.memocount * sizeof(Memo);
+	for (i = 0; i < ni->memos.memocount; i++) {
+	    count++;
+	    if (ni->memos.memos[i].text)
+		mem += strlen(ni->memos.memos[i].text)+1;
+	}
+    }
+    for (ci = cs_firstchan(); ci; ci = cs_nextchan()) {
+	mem += ci->memos.memocount * sizeof(Memo);
+	for (i = 0; i < ci->memos.memocount; i++) {
+	    if (ci->memos.memos[i].text)
+		mem += strlen(ci->memos.memos[i].text)+1;
+	}
+    }
+    *nrec = count;
+    *memuse = mem;
+}
+
+/*************************************************************************/
+
 /* Compatibility memo load routine. */
 
 #define SAFE(x) do {					\
@@ -165,6 +212,8 @@ void load_old_ms_dbase(void)
 	    }
 	}
 	break;
+      case -1:
+	fatal("Unable to read version number from memo.db");
       default:
 	fatal("Unsupported version number (%d) on memo.db", ver);
     } /* switch (version) */
@@ -180,11 +229,10 @@ void load_old_ms_dbase(void)
 
 void check_memos(User *u)
 {
-    NickInfo *ni;
+    NickInfo *ni = u->ni;
     int i, newcnt = 0;
 
-    if (!(ni = u->ni) || !nick_recognized(u) ||
-			 !(ni->flags & NI_MEMO_SIGNON))
+    if (!ni || !nick_recognized(u) || !(ni->flags & NI_MEMO_SIGNON))
 	return;
 
     for (i = 0; i < ni->memos.memocount; i++) {
@@ -221,153 +269,59 @@ void check_memos(User *u)
 
 /* Return the MemoInfo corresponding to the given nick or channel name.
  * Return in `ischan' 1 if the name was a channel name, else 0.
+ * Return in `error' a GMI_* error code if the return value is NULL.
+ * Also set `error' to GMI_SUSPENDED if the nick/channel is suspended,
+ * even though a valid MemoInfo is returned.
  */
 
-static MemoInfo *getmemoinfo(const char *name, int *ischan)
+static MemoInfo *getmemoinfo(const char *name, int *ischan, int *error)
 {
+    if (error)
+	*error = 0;
+
     if (*name == '#') {
 	ChannelInfo *ci;
 	if (ischan)
 	    *ischan = 1;
 	ci = cs_findchan(name);
-	if (ci)
-	    return &ci->memos;
-	else
+
+	if (ci) {
+	    if (ci->flags & CI_VERBOTEN) {
+		if (error)
+		    *error = GMI_FORBIDDEN;
+		return NULL;
+	    } else {
+		if (ci->suspendinfo && error)
+		    *error = GMI_SUSPENDED;
+		return &ci->memos;
+	    }
+	} else {
+	    if (error)
+		*error = GMI_NOTFOUND;
 	    return NULL;
+	}
+
     } else {
 	NickInfo *ni;
 	if (ischan)
 	    *ischan = 0;
 	ni = findnick(name);
-	if (ni)
-	    return &getlink(ni)->memos;
-	else
-	    return NULL;
-    }
-}
-
-/*************************************************************************/
-
-/* Delete a memo by number.  Return 1 if the memo was found, else 0. */
-
-static int delmemo(MemoInfo *mi, int num)
-{
-    int i;
-
-    for (i = 0; i < mi->memocount; i++) {
-	if (mi->memos[i].number == num)
-	    break;
-    }
-    if (i < mi->memocount) {
-	free(mi->memos[i].text); /* Deallocate memo text memory */
-	mi->memocount--;	 /* One less memo now */
-	if (i < mi->memocount)	 /* Move remaining memos down a slot */
-	    memmove(mi->memos + i, mi->memos + i+1,
-				sizeof(Memo) * (mi->memocount - i));
-	if (mi->memocount == 0) { /* If no more memos, free array */
-	    free(mi->memos);
-	    mi->memos = NULL;
-	}
-	return 1;
-    } else {
-	return 0;
-    }
-}
-
-/*************************************************************************/
-/*********************** MemoServ command routines ***********************/
-/*************************************************************************/
-
-/* Return a help message. */
-
-static void do_help(User *u)
-{
-    char *cmd = strtok(NULL, "");
-
-    if (!cmd) {
-	notice_help(s_MemoServ, u, MEMO_HELP, s_ChanServ);
-    } else {
-	help_cmd(s_MemoServ, u, cmds, cmd);
-    }
-}
-
-/*************************************************************************/
-
-/* Send a memo to a nick/channel. */
-
-static void do_send(User *u)
-{
-    char *source = u->nick;
-    int ischan;
-    MemoInfo *mi;
-    Memo *m;
-    char *name = strtok(NULL, " ");
-    char *text = strtok(NULL, "");
-    time_t now = time(NULL);
-    int is_servadmin = is_services_admin(u);
-
-    if (!text) {
-	syntax_error(s_MemoServ, u, "SEND", MEMO_SEND_SYNTAX);
-
-    } else if (!nick_recognized(u)) {
-	notice_lang(s_MemoServ, u, NICK_IDENTIFY_REQUIRED, s_NickServ);
-
-    } else if (!(mi = getmemoinfo(name, &ischan))) {
-	notice_lang(s_MemoServ, u,
-		ischan ? CHAN_X_NOT_REGISTERED : NICK_X_NOT_REGISTERED, name);
-
-    } else if (MSSendDelay > 0 &&
-		u && u->lastmemosend+MSSendDelay > now && !is_servadmin) {
-	u->lastmemosend = now;
-	notice_lang(s_MemoServ, u, MEMO_SEND_PLEASE_WAIT, MSSendDelay);
-
-    } else if (mi->memomax == 0 && !is_servadmin) {
-	notice_lang(s_MemoServ, u, MEMO_X_GETS_NO_MEMOS, name);
-
-    } else if (mi->memomax > 0 && mi->memocount >= mi->memomax
-		&& !is_servadmin) {
-	notice_lang(s_MemoServ, u, MEMO_X_HAS_TOO_MANY_MEMOS, name);
-
-    } else {
-	u->lastmemosend = now;
-	mi->memocount++;
-	mi->memos = srealloc(mi->memos, sizeof(Memo) * mi->memocount);
-	m = &mi->memos[mi->memocount-1];
-	strscpy(m->sender, source, NICKMAX);
-	if (mi->memocount > 1) {
-	    m->number = m[-1].number + 1;
-	    if (m->number < 1) {
-		int i;
-		for (i = 0; i < mi->memocount; i++)
-		    mi->memos[i].number = i+1;
+	if (ni) {
+	    if (ni->status & NS_VERBOTEN) {
+		if (error)
+		    *error = GMI_FORBIDDEN;
+		return NULL;
+	    } else {
+		if (ni->suspendinfo && error)
+		    *error = GMI_SUSPENDED;
+		return &getlink(ni)->memos;
 	    }
 	} else {
-	    m->number = 1;
+	    if (error)
+		*error = GMI_NOTFOUND;
+	    return NULL;
 	}
-	m->time = time(NULL);
-	m->text = sstrdup(text);
-	m->flags = MF_UNREAD;
-	notice_lang(s_MemoServ, u, MEMO_SENT, name);
-	if (!ischan) {
-	    NickInfo *ni = getlink(findnick(name));
-	    if (ni->flags & NI_MEMO_RECEIVE) {
-		if (MSNotifyAll) {
-		    for (u = firstuser(); u; u = nextuser()) {
-			if (u->real_ni == ni) {
-			    notice_lang(s_MemoServ, u, MEMO_NEW_MEMO_ARRIVED,
-					source, s_MemoServ, m->number);
-			}
-		    }
-		} else {
-		    u = finduser(name);
-		    if (u) {
-			notice_lang(s_MemoServ, u, MEMO_NEW_MEMO_ARRIVED,
-					source, s_MemoServ, m->number);
-		    }
-		} /* if (MSNotifyAll) */
-	    } /* if (flags & MEMO_RECEIVE) */
-	} /* if (!ischan) */
-    } /* if command is valid */
+    }
 }
 
 /*************************************************************************/
@@ -407,6 +361,8 @@ static int list_memo(User *u, int index, MemoInfo *mi, int *sent_header,
     return 1;
 }
 
+/* List callback. */
+
 static int list_memo_callback(User *u, int num, va_list args)
 {
     MemoInfo *mi = va_arg(args, MemoInfo *);
@@ -422,8 +378,201 @@ static int list_memo_callback(User *u, int num, va_list args)
     return list_memo(u, i, mi, sent_header, 0, chan);
 }
 
+/*************************************************************************/
 
-/* List the memos (if any) for the source nick or given channel. */
+/* Send a single memo to the given user. */
+
+static int read_memo(User *u, int index, MemoInfo *mi, const char *chan)
+{
+    Memo *m;
+    char timebuf[64];
+    struct tm tm;
+
+    if (index < 0 || index >= mi->memocount)
+	return 0;
+    m = &mi->memos[index];
+    tm = *localtime(&m->time);
+    strftime_lang(timebuf, sizeof(timebuf),
+		u, STRFTIME_DATE_TIME_FORMAT, &tm);
+    timebuf[sizeof(timebuf)-1] = 0;
+    if (chan)
+	notice_lang(s_MemoServ, u, MEMO_CHAN_HEADER, m->number,
+		m->sender, timebuf, s_MemoServ, chan, m->number);
+    else
+	notice_lang(s_MemoServ, u, MEMO_HEADER, m->number,
+		m->sender, timebuf, s_MemoServ, m->number);
+    notice_lang(s_MemoServ, u, MEMO_TEXT, m->text);
+    m->flags &= ~MF_UNREAD;
+    return 1;
+}
+
+/* List callback. */
+
+static int read_memo_callback(User *u, int num, va_list args)
+{
+    MemoInfo *mi = va_arg(args, MemoInfo *);
+    const char *chan = va_arg(args, const char *);
+    int i;
+
+    for (i = 0; i < mi->memocount; i++) {
+	if (mi->memos[i].number == num)
+	    break;
+    }
+    /* Range check done in read_memo */
+    return read_memo(u, i, mi, chan);
+}
+
+/*************************************************************************/
+
+/* Delete a memo by number.  Return 1 if the memo was found, else 0. */
+
+static int del_memo(MemoInfo *mi, int num)
+{
+    int i;
+
+    for (i = 0; i < mi->memocount; i++) {
+	if (mi->memos[i].number == num)
+	    break;
+    }
+    if (i < mi->memocount) {
+	free(mi->memos[i].text); /* Deallocate memo text memory */
+	mi->memocount--;	 /* One less memo now */
+	if (i < mi->memocount)	 /* Move remaining memos down a slot */
+	    memmove(mi->memos + i, mi->memos + i+1,
+				sizeof(Memo) * (mi->memocount - i));
+	if (mi->memocount == 0) { /* If no more memos, free array */
+	    free(mi->memos);
+	    mi->memos = NULL;
+	}
+	return 1;
+    } else {
+	return 0;
+    }
+}
+
+/* Delete a single memo from a MemoInfo. */
+
+static int del_memo_callback(User *u, int num, va_list args)
+{
+    MemoInfo *mi = va_arg(args, MemoInfo *);
+    int *last = va_arg(args, int *);
+
+    if (del_memo(mi, num)) {
+	*last = num;
+	return 1;
+    } else {
+	return 0;
+    }
+}
+
+/*************************************************************************/
+/*********************** MemoServ command routines ***********************/
+/*************************************************************************/
+
+/* Return a help message. */
+
+static void do_help(User *u)
+{
+    char *cmd = strtok(NULL, "");
+
+    if (!cmd) {
+	notice_help(s_MemoServ, u, MEMO_HELP, s_ChanServ);
+    } else {
+	help_cmd(s_MemoServ, u, cmds, cmd);
+    }
+}
+
+/*************************************************************************/
+
+/* Send a memo to a nick/channel. */
+
+static void do_send(User *u)
+{
+    char *source = u->nick;
+    int ischan, error;
+    MemoInfo *mi;
+    Memo *m;
+    char *name = strtok(NULL, " ");
+    char *text = strtok(NULL, "");
+    time_t now = time(NULL);
+    int is_servadmin = is_services_admin(u);
+
+    if (!text) {
+	syntax_error(s_MemoServ, u, "SEND", MEMO_SEND_SYNTAX);
+
+    } else if (!nick_recognized(u)) {
+	notice_lang(s_MemoServ, u, NICK_IDENTIFY_REQUIRED, s_NickServ);
+
+    } else if (!(mi = getmemoinfo(name, &ischan, &error))) {
+	if (error == GMI_FORBIDDEN)
+	    notice_lang(s_MemoServ, u,
+		ischan ? CHAN_X_FORBIDDEN: NICK_X_FORBIDDEN,
+		name);
+	else
+	    notice_lang(s_MemoServ, u,
+		ischan ? CHAN_X_NOT_REGISTERED : NICK_X_NOT_REGISTERED, name);
+
+    } else if (error == GMI_SUSPENDED) {
+	notice_lang(s_MemoServ, u,
+	    ischan ? CHAN_X_SUSPENDED_MEMOS : NICK_X_SUSPENDED_MEMOS, name);
+
+    } else if (MSSendDelay > 0 &&
+		u && u->lastmemosend+MSSendDelay > now && !is_servadmin) {
+	u->lastmemosend = now;
+	notice_lang(s_MemoServ, u, MEMO_SEND_PLEASE_WAIT, MSSendDelay);
+
+    } else if (mi->memomax == 0 && !is_servadmin) {
+	notice_lang(s_MemoServ, u, MEMO_X_GETS_NO_MEMOS, name);
+
+    } else if (mi->memomax > 0 && mi->memocount >= mi->memomax
+		&& !is_servadmin) {
+	notice_lang(s_MemoServ, u, MEMO_X_HAS_TOO_MANY_MEMOS, name);
+
+    } else {
+	u->lastmemosend = now;
+	mi->memocount++;
+	mi->memos = srealloc(mi->memos, sizeof(Memo) * mi->memocount);
+	m = &mi->memos[mi->memocount-1];
+	strscpy(m->sender, source, NICKMAX);
+	if (mi->memocount > 1) {
+	    m->number = m[-1].number + 1;
+	    if (m->number < 1) {
+		int i;
+		for (i = 0; i < mi->memocount; i++)
+		    mi->memos[i].number = i+1;
+	    }
+	} else {
+	    m->number = 1;
+	}
+	m->time = time(NULL);
+	m->text = sstrdup(text);
+	m->flags = MF_UNREAD;
+	notice_lang(s_MemoServ, u, MEMO_SENT, name);
+	if (!ischan) {
+	    NickInfo *ni = getlink(findnick(name));  /* cannot fail */
+	    if (ni->flags & NI_MEMO_RECEIVE) {
+		if (MSNotifyAll) {
+		    for (u = firstuser(); u; u = nextuser()) {
+			if (u->ni == ni) {
+			    notice_lang(s_MemoServ, u, MEMO_NEW_MEMO_ARRIVED,
+					source, s_MemoServ, m->number);
+			}
+		    }
+		} else {
+		    u = finduser(name);
+		    if (u && (u->real_ni->status & NS_RECOGNIZED)) {
+			notice_lang(s_MemoServ, u, MEMO_NEW_MEMO_ARRIVED,
+					source, s_MemoServ, m->number);
+		    }
+		} /* if (MSNotifyAll) */
+	    } /* if (flags & MEMO_RECEIVE) */
+	} /* if (!ischan) */
+    } /* if command is valid */
+}
+
+/*************************************************************************/
+
+/* List memos for the source nick or given channel. */
 
 static void do_list(User *u)
 {
@@ -441,7 +590,7 @@ static void do_list(User *u)
 	    return;
 	} else if (ci->flags & CI_VERBOTEN) {
 	    notice_lang(s_MemoServ, u, CHAN_X_FORBIDDEN, chan);
-	    return;            
+	    return;
 	} else if (!check_access(u, ci, CA_MEMO)) {
 	    notice_lang(s_MemoServ, u, ACCESS_DENIED);
 	    return;
@@ -492,47 +641,6 @@ static void do_list(User *u)
 
 /*************************************************************************/
 
-/* Send a single memo to the given user. */
-
-static int read_memo(User *u, int index, MemoInfo *mi, const char *chan)
-{
-    Memo *m;
-    char timebuf[64];
-    struct tm tm;
-
-    if (index < 0 || index >= mi->memocount)
-	return 0;
-    m = &mi->memos[index];
-    tm = *localtime(&m->time);
-    strftime_lang(timebuf, sizeof(timebuf),
-		u, STRFTIME_DATE_TIME_FORMAT, &tm);
-    timebuf[sizeof(timebuf)-1] = 0;
-    if (chan)
-	notice_lang(s_MemoServ, u, MEMO_CHAN_HEADER, m->number,
-		m->sender, timebuf, s_MemoServ, chan, m->number);
-    else
-	notice_lang(s_MemoServ, u, MEMO_HEADER, m->number,
-		m->sender, timebuf, s_MemoServ, m->number);
-    notice_lang(s_MemoServ, u, MEMO_TEXT, m->text);
-    m->flags &= ~MF_UNREAD;
-    return 1;
-}
-
-static int read_memo_callback(User *u, int num, va_list args)
-{
-    MemoInfo *mi = va_arg(args, MemoInfo *);
-    const char *chan = va_arg(args, const char *);
-    int i;
-
-    for (i = 0; i < mi->memocount; i++) {
-	if (mi->memos[i].number == num)
-	    break;
-    }
-    /* Range check done in read_memo */
-    return read_memo(u, i, mi, chan);
-}
-
-
 /* Read memos. */
 
 static void do_read(User *u)
@@ -610,37 +718,6 @@ static void do_read(User *u)
 
 /*************************************************************************/
 
-/* Delete a single memo from a MemoInfo. */
-
-static int del_memo_callback(User *u, int num, va_list args)
-{
-    MemoInfo *mi = va_arg(args, MemoInfo *);
-    int *last = va_arg(args, int *);
-    int *last0 = va_arg(args, int *);
-    char **end = va_arg(args, char **);
-    int *left = va_arg(args, int *);
-
-    if (delmemo(mi, num)) {
-	if (num != (*last)+1) {
-	    if (*last != -1) {
-		int len;
-		if (*last0 != *last)
-		    len = snprintf(*end, *left, ",%d-%d", *last0, *last);
-		else
-		    len = snprintf(*end, *left, ",%d", *last);
-		*end += len;
-		*left -= len;
-	    }
-	    *last0 = num;
-	}
-	*last = num;
-	return 1;
-    } else {
-	return 0;
-    }
-}
-
-
 /* Delete memos. */
 
 static void do_del(User *u)
@@ -648,9 +725,8 @@ static void do_del(User *u)
     MemoInfo *mi;
     ChannelInfo *ci;
     char *numstr = strtok(NULL, ""), *chan = NULL;
-    int last, last0, i;
-    char buf[BUFSIZE], *end;
-    int delcount, count, left;
+    int last, i;
+    int delcount, count;
 
     if (numstr && *numstr == '#') {
 	chan = strtok(numstr, " ");
@@ -683,28 +759,16 @@ static void do_del(User *u)
     } else {
 	if (isdigit(*numstr)) {
 	    /* Delete a specific memo or memos. */
-	    last = -1;   /* Last memo deleted */
-	    last0 = -1;  /* Beginning of range of last memos deleted */
-	    end = buf;
-	    left = sizeof(buf);
-	    delcount = process_numlist(numstr, &count, del_memo_callback, u, mi,
-						&last, &last0, &end, &left);
-	    if (last != -1) {
-		/* Some memos got deleted; tell them which ones. */
-		if (delcount > 1) {
-		    if (last0 != last)
-			end += snprintf(end, sizeof(buf)-(end-buf),
-				",%d-%d", last0, last);
-		    else
-			end += snprintf(end, sizeof(buf)-(end-buf),
-				",%d", last);
-		    /* "buf+1" here because *buf == ',' */
-		    notice_lang(s_MemoServ, u, MEMO_DELETED_SEVERAL, buf+1);
-		} else {
+	    delcount = process_numlist(numstr, &count, del_memo_callback,
+				       u, mi, &last);
+	    if (delcount) {
+		/* Some memos got deleted. */
+		if (delcount > 1)
+		    notice_lang(s_MemoServ, u, MEMO_DELETED_SEVERAL, delcount);
+		else
 		    notice_lang(s_MemoServ, u, MEMO_DELETED_ONE, last);
-		}
 	    } else {
-		/* No memos were deleted.  Tell them so. */
+		/* No memos were deleted. */
 		if (count == 1)
 		    notice_lang(s_MemoServ, u, MEMO_DOES_NOT_EXIST,
 				atoi(numstr));
@@ -775,11 +839,14 @@ static void do_set_notify(User *u, MemoInfo *mi, char *param)
 
 /*************************************************************************/
 
+/* Regular user parameters: [#channel] number
+ * Services admin parameters: [#channel | nick] {number | NONE} [HARD]
+ */
+
 static void do_set_limit(User *u, MemoInfo *mi, char *param)
 {
     char *p1 = strtok(param, " ");
     char *p2 = strtok(NULL, " ");
-    char *p3 = strtok(NULL, " ");
     char *user = NULL, *chan = NULL;
     int32 limit;
     NickInfo *ni = u->ni;
@@ -789,14 +856,13 @@ static void do_set_limit(User *u, MemoInfo *mi, char *param)
     if (p1 && *p1 == '#') {
 	chan = p1;
 	p1 = p2;
-	p2 = p3;
-	p3 = strtok(NULL, " ");
+	p2 = strtok(NULL, " ");
 	if (!(ci = cs_findchan(chan))) {
 	    notice_lang(s_MemoServ, u, CHAN_X_NOT_REGISTERED, chan);
 	    return;
 	} else if (ci->flags & CI_VERBOTEN) {
 	    notice_lang(s_MemoServ, u, CHAN_X_FORBIDDEN, chan);
-	    return;            
+	    return;
 	} else if (!is_servadmin && !check_access(u, ci, CA_MEMO)) {
 	    notice_lang(s_MemoServ, u, ACCESS_DENIED);
 	    return;
@@ -813,7 +879,7 @@ static void do_set_limit(User *u, MemoInfo *mi, char *param)
 	    user = p1;
 	    mi = &ni->memos;
 	    p1 = p2;
-	    p2 = p3;
+	    p2 = strtok(NULL, " ");
 	} else if (!p1) {
 	    syntax_error(s_MemoServ, u, "SET LIMIT",
 					MEMO_SET_LIMIT_SERVADMIN_SYNTAX);
@@ -908,6 +974,9 @@ static void do_info(User *u)
 	if (!ni) {
 	    notice_lang(s_MemoServ, u, NICK_X_NOT_REGISTERED, name);
 	    return;
+	} else if (ni->status & NS_VERBOTEN) {
+	    notice_lang(s_MemoServ, u, NICK_X_FORBIDDEN, name);
+	    return;
 	}
 	ni = getlink(ni);
 	mi = &ni->memos;
@@ -918,8 +987,8 @@ static void do_info(User *u)
 	    notice_lang(s_MemoServ, u, CHAN_X_NOT_REGISTERED, name);
 	    return;
 	} else if (ci->flags & CI_VERBOTEN) {
-	    notice_lang(s_MemoServ, u, CHAN_X_FORBIDDEN, chan);
-	    return;            
+	    notice_lang(s_MemoServ, u, CHAN_X_FORBIDDEN, name);
+	    return;
 	} else if (!check_access(u, ci, CA_MEMO)) {
 	    notice_lang(s_MemoServ, u, ACCESS_DENIED);
 	    return;
@@ -931,11 +1000,11 @@ static void do_info(User *u)
 	    notice_lang(s_MemoServ, u, NICK_IDENTIFY_REQUIRED, s_NickServ);
 	    return;
 	}
+	ni = u->ni;
 	mi = &u->ni->memos;
     }
 
     if (name && ni != u->ni) {
-
 	if (!mi->memocount) {
 	    notice_lang(s_MemoServ, u, MEMO_INFO_X_NO_MEMOS, name);
 	} else if (mi->memocount == 1) {
@@ -963,6 +1032,7 @@ static void do_info(User *u)
 			name, mi->memocount, count);
 	}
 	if (mi->memomax >= 0) {
+	    notice(s_MemoServ, u->nick, "hardmax: %d", hardmax);
 	    if (hardmax)
 		notice_lang(s_MemoServ, u, MEMO_INFO_X_HARD_LIMIT,
 			name, mi->memomax);
@@ -972,6 +1042,17 @@ static void do_info(User *u)
 	} else {
 	    notice_lang(s_MemoServ, u, MEMO_INFO_X_NO_LIMIT, name);
 	}
+	if (ni) {	/* only nicks are notified of new memos */
+	    if ((ni->flags & NI_MEMO_RECEIVE) && (ni->flags & NI_MEMO_SIGNON)) {
+		notice_lang(s_MemoServ, u, MEMO_INFO_X_NOTIFY_ON, name);
+	    } else if (ni->flags & NI_MEMO_RECEIVE) {
+		notice_lang(s_MemoServ, u, MEMO_INFO_X_NOTIFY_RECEIVE, name);
+	    } else if (ni->flags & NI_MEMO_SIGNON) {
+		notice_lang(s_MemoServ, u, MEMO_INFO_X_NOTIFY_SIGNON, name);
+	    } else {
+		notice_lang(s_MemoServ, u, MEMO_INFO_X_NOTIFY_OFF, name);
+	    }
+	}
 
     } else { /* !name || ni == u->ni */
 
@@ -1012,6 +1093,15 @@ static void do_info(User *u)
 	} else {
 	    notice_lang(s_MemoServ, u, MEMO_INFO_NO_LIMIT);
 	}
+	if ((ni->flags & NI_MEMO_RECEIVE) && (ni->flags & NI_MEMO_SIGNON)) {
+	    notice_lang(s_MemoServ, u, MEMO_INFO_NOTIFY_ON);
+	} else if (ni->flags & NI_MEMO_RECEIVE) {
+	    notice_lang(s_MemoServ, u, MEMO_INFO_NOTIFY_RECEIVE);
+	} else if (ni->flags & NI_MEMO_SIGNON) {
+	    notice_lang(s_MemoServ, u, MEMO_INFO_NOTIFY_SIGNON);
+	} else {
+	    notice_lang(s_MemoServ, u, MEMO_INFO_NOTIFY_OFF);
+	}
 
     } /* if (name && ni != u->ni) */
 }
