@@ -1,22 +1,38 @@
 /* OperServ functions.
  *
- * Services is copyright (c) 1996-1999 Andy Church.
- *     E-mail: <achurch@dragonfire.net>
+ * IRC Services is copyright (c) 1996-2001 Andrew Church.
+ *     E-mail: <achurch@achurch.org>
+ * Parts copyright (c) 1999-2000 Andrew Kempe and others.
  * This program is free but copyrighted software; see the file COPYING for
  * details.
  */
 
 #include "services.h"
 #include "pseudo.h"
+#include "news.h"
+
+#include <unistd.h>
 
 /*************************************************************************/
 
+/* Message to send when unable to set modes or kick on a channel. */
+#if defined(IRC_DALNET) || defined(IRC_UNDERNET)
+# define OPER_BOUNCY_MODES_MESSAGE OPER_BOUNCY_MODES_U_LINE
+#else
+# define OPER_BOUNCY_MODES_MESSAGE OPER_BOUNCY_MODES
+#endif
+
 /* Services admin list */
 static NickInfo *services_admins[MAX_SERVADMINS];
 
 /* Services operator list */
 static NickInfo *services_opers[MAX_SERVOPERS];
 
+/* Services super-user password */
+static char supass[PASSMAX];
+/* Is the password currently unset? */
+static char no_supass = 1;
+
 
 struct clone {
     char *host;
@@ -39,7 +55,9 @@ static void do_admin(User *u);
 static void do_oper(User *u);
 static void do_os_mode(User *u);
 static void do_clearmodes(User *u);
+static void do_clearchan(User *u);
 static void do_os_kick(User *u);
+static void do_su(User *u);
 static void do_set(User *u);
 static void do_jupe(User *u);
 static void do_raw(User *u);
@@ -69,25 +87,38 @@ static Command cmds[] = {
     { "ADMIN",      do_admin,      NULL,  OPER_HELP_ADMIN,      -1,-1,-1,-1 },
     { "OPER",       do_oper,       NULL,  OPER_HELP_OPER,       -1,-1,-1,-1 },
     /* Similarly, anyone can use *NEWS LIST, but *NEWS {ADD,DEL} are
-     * reserved for Services admins. */
+     * reserved for Services operators. */
     { "LOGONNEWS",  do_logonnews,  NULL,  NEWS_HELP_LOGON,      -1,-1,-1,-1 },
     { "OPERNEWS",   do_opernews,   NULL,  NEWS_HELP_OPER,       -1,-1,-1,-1 },
 
     /* Commands for Services opers: */
     { "MODE",       do_os_mode,    is_services_oper,
 	OPER_HELP_MODE, -1,-1,-1,-1 },
-    { "CLEARMODES", do_clearmodes, is_services_oper,
-	OPER_HELP_CLEARMODES, -1,-1,-1,-1 },
     { "KICK",       do_os_kick,    is_services_oper,
 	OPER_HELP_KICK, -1,-1,-1,-1 },
+    { "CLEARMODES", do_clearmodes, is_services_oper,
+	OPER_HELP_CLEARMODES, -1,-1,-1,-1 },
+    { "CLEARCHAN",  do_clearchan,  is_services_oper,
+	OPER_HELP_CLEARCHAN, -1,-1,-1,-1 },
+    { "KILLCLONES", do_killclones, is_services_oper,
+	OPER_HELP_KILLCLONES, -1,-1,-1, -1 },
     { "AKILL",      do_akill,      is_services_oper,
 	OPER_HELP_AKILL, -1,-1,-1,-1 },
+#ifndef STREAMLINED
+    { "SESSION",    do_session,    is_services_oper,
+        OPER_HELP_SESSION, -1,-1,-1, -1 },
+    { "EXCEPTION",  do_exception,  is_services_oper,
+        OPER_HELP_EXCEPTION, -1,-1,-1, -1 },
+#endif
 
     /* Commands for Services admins: */
+    { "SU",         do_su,         NULL,  /* do_su() checks perms itself */
+	OPER_HELP_SU, -1,-1,-1,-1 },
     { "SET",        do_set,        is_services_admin,
 	OPER_HELP_SET, -1,-1,-1,-1 },
     { "SET READONLY",0,0,  OPER_HELP_SET_READONLY, -1,-1,-1,-1 },
     { "SET DEBUG",0,0,     OPER_HELP_SET_DEBUG, -1,-1,-1,-1 },
+    { "SET SUPASS",0,0,    OPER_HELP_SET_SUPASS, -1,-1,-1,-1 },
     { "JUPE",       do_jupe,       is_services_admin,
 	OPER_HELP_JUPE, -1,-1,-1,-1 },
     { "RAW",        do_raw,        is_services_admin,
@@ -102,21 +133,11 @@ static Command cmds[] = {
 	OPER_HELP_RESTART, -1,-1,-1,-1 },
     { "LISTIGNORE", do_listignore, is_services_admin,
 	-1,-1,-1,-1, -1 },
-    { "KILLCLONES", do_killclones, is_services_admin,
-	OPER_HELP_KILLCLONES, -1,-1,-1, -1 },
-#ifndef STREAMLINED
-    { "SESSION",    do_session,    is_services_admin,
-        OPER_HELP_SESSION, -1,-1,-1, -1 },
-    { "EXCEPTION",  do_exception,  is_services_admin,
-        OPER_HELP_EXCEPTION, -1,-1,-1, -1 },
-#endif
 
-    /* Commands for Services root: */
-
-    { "ROTATELOG",  rotate_log,  is_services_root, -1,-1,-1,-1,
-	OPER_HELP_ROTATELOG },
+    /* Commands for Services super-user: */
 
 #ifdef DEBUG_COMMANDS
+    { "LISTSERVERS",send_server_list,   is_services_root, -1,-1,-1,-1,-1 },
     { "LISTCHANS",  send_channel_list,  is_services_root, -1,-1,-1,-1,-1 },
     { "LISTCHAN",   send_channel_users, is_services_root, -1,-1,-1,-1,-1 },
     { "LISTUSERS",  send_user_list,     is_services_root, -1,-1,-1,-1,-1 },
@@ -182,7 +203,7 @@ void operserv(const char *source, char *buf)
 }
 
 /*************************************************************************/
-/**************************** Privilege checks ***************************/
+/************************ Database loading/saving ************************/
 /*************************************************************************/
 
 /* Load OperServ data. */
@@ -192,34 +213,39 @@ void operserv(const char *source, char *buf)
 	if (!forceload)					\
 	    fatal("Read error on %s", OperDBName);	\
 	failed = 1;					\
-	break;						\
+	close_db(f);					\
+	return;						\
     }							\
 } while (0)
 
 void load_os_dbase(void)
 {
     dbFILE *f;
-    int16 i, n, ver;
+    int16 i, n;
+    int32 ver;
     char *s;
     int failed = 0;
 
     if (!(f = open_db(s_OperServ, OperDBName, "r")))
 	return;
     switch (ver = get_file_version(f)) {
+      case 11:
+      case 10:
+      case 9:
+      case 8:
       case 7:
       case 6:
       case 5:
 	SAFE(read_int16(&n, f));
-	for (i = 0; i < n && !failed; i++) {
+	for (i = 0; i < n; i++) {
 	    SAFE(read_string(&s, f));
 	    if (s && i < MAX_SERVADMINS)
 		services_admins[i] = findnick(s);
 	    if (s)
 		free(s);
 	}
-	if (!failed)
-	    SAFE(read_int16(&n, f));
-	for (i = 0; i < n && !failed; i++) {
+	SAFE(read_int16(&n, f));
+	for (i = 0; i < n; i++) {
 	    SAFE(read_string(&s, f));
 	    if (s && i < MAX_SERVOPERS)
 		services_opers[i] = findnick(s);
@@ -232,12 +258,17 @@ void load_os_dbase(void)
 	    SAFE(read_int32(&tmp32, f));
 	    maxusertime = tmp32;
 	}
+	if (ver >= 9) {
+	    SAFE(read_int8(&no_supass, f));
+	    if (!no_supass)
+		SAFE(read_buffer(supass, f));
+	}
 	break;
 
       case 4:
       case 3:
 	SAFE(read_int16(&n, f));
-	for (i = 0; i < n && !failed; i++) {
+	for (i = 0; i < n; i++) {
 	    SAFE(read_string(&s, f));
 	    if (s && i < MAX_SERVADMINS)
 		services_admins[i] = findnick(s);
@@ -246,6 +277,9 @@ void load_os_dbase(void)
 	}
 	break;
 
+      case -1:
+	fatal("Unable to read version number from %s", OperDBName);
+
       default:
 	fatal("Unsupported version (%d) on %s", ver, OperDBName);
     } /* switch (version) */
@@ -258,18 +292,7 @@ void load_os_dbase(void)
 
 /* Save OperServ data. */
 
-#define SAFE(x) do {						\
-    if ((x) < 0) {						\
-	restore_db(f);						\
-	log_perror("Write error on %s", OperDBName);		\
-	if (time(NULL) - lastwarn > WarningTimeout) {		\
-	    wallops(NULL, "Write error on %s: %s", OperDBName,	\
-			strerror(errno));			\
-	    lastwarn = time(NULL);				\
-	}							\
-	return;							\
-    }								\
-} while (0)
+#define SAFE(x) do { if ((x) < 0) goto fail; } while (0)
 
 void save_os_dbase(void)
 {
@@ -300,18 +323,35 @@ void save_os_dbase(void)
     }
     SAFE(write_int32(maxusercnt, f));
     SAFE(write_int32(maxusertime, f));
+    SAFE(write_int8(no_supass, f));
+    if (!no_supass)
+	SAFE(write_buffer(supass, f));
     close_db(f);
+    return;
+
+  fail:
+    restore_db(f);
+    log_perror("Write error on %s", OperDBName);
+    if (time(NULL) - lastwarn > WarningTimeout) {
+	wallops(NULL, "Write error on %s: %s", OperDBName,
+		strerror(errno));
+	lastwarn = time(NULL);
+    }
 }
 
 #undef SAFE
 
+/*************************************************************************/
+/**************************** Privilege checks ***************************/
 /*************************************************************************/
 
-/* Does the given user have Services root privileges? */
+/* Does the given user have Services super-user privileges? */
 
 int is_services_root(User *u)
 {
-    if (!(u->mode & UMODE_O) || stricmp(u->nick, ServicesRoot) != 0)
+    if (u->real_ni && (u->real_ni->status & NS_SERVROOT))
+	return 1;
+    if (!is_oper_u(u) || stricmp(u->nick, ServicesRoot) != 0)
 	return 0;
     if (skeleton || nick_identified(u))
 	return 1;
@@ -326,7 +366,7 @@ int is_services_admin(User *u)
 {
     int i;
 
-    if (!(u->mode & UMODE_O))
+    if (!is_oper_u(u))
 	return 0;
     if (is_services_root(u))
 	return 1;
@@ -350,7 +390,7 @@ int is_services_oper(User *u)
 {
     int i;
 
-    if (!(u->mode & UMODE_O))
+    if (!is_oper_u(u))
 	return 0;
     if (is_services_admin(u))
 	return 1;
@@ -370,9 +410,9 @@ int is_services_oper(User *u)
 
 /* Is the given nick a Services admin/root nick? */
 
-/* NOTE: Do not use this to check if a user who is online is a services admin
- * or root. This function only checks if a user has the ABILITY to be a 
- * services admin. Rather use is_services_admin(User *u). -TheShadow */
+/* NOTE: Do not use this to check if a user who is online is a services
+ * admin or root. This function only checks if a user has the ABILITY to be
+ * a services admin. Rather use is_services_admin(User *u). -TheShadow */
 
 int nick_is_services_admin(NickInfo *ni)
 {
@@ -391,7 +431,6 @@ int nick_is_services_admin(NickInfo *ni)
 
 /*************************************************************************/
 
-
 /* Expunge a deleted nick from the Services admin/oper lists. */
 
 void os_remove_nick(const NickInfo *ni)
@@ -476,6 +515,9 @@ void check_clones(User *user)
 
 static void send_clone_lists(User *u)
 {
+#ifdef STREAMLINED
+    notice(s_OperServ, u->nick, "No clone checking in STREAMLINED mode.");
+#else
     int i;
 
     if (!CheckClones) {
@@ -493,10 +535,101 @@ static void send_clone_lists(User *u)
 	if (clonelist[i].host)
 	    notice(s_OperServ, u->nick, "    %10ld  %s", warnings[i].time, warnings[i].host ? warnings[i].host : "(null)");
     }
+#endif /* !STREAMLINED */
 }
 
 #endif	/* DEBUG_COMMANDS */
 
+/*************************************************************************/
+/********************** Admin/oper list modification *********************/
+/*************************************************************************/
+
+#define LIST_ADMIN	0
+#define LIST_OPER	1
+
+#define MSG_EXISTS	0
+#define MSG_ADDED	1
+#define MSG_TOOMANY	2
+#define MSG_REMOVED	3
+#define MSG_NOTFOUND	4
+
+static int privlist_msgs[2][5] = {
+    { OPER_ADMIN_EXISTS,
+      OPER_ADMIN_ADDED,
+      OPER_ADMIN_TOO_MANY,
+      OPER_ADMIN_REMOVED,
+      OPER_ADMIN_NOT_FOUND,
+    },
+    { OPER_OPER_EXISTS,
+      OPER_OPER_ADDED,
+      OPER_OPER_TOO_MANY,
+      OPER_OPER_REMOVED,
+      OPER_OPER_NOT_FOUND,
+    },
+};
+
+/*************************************************************************/
+
+/* Add a nick to the Services admin/oper list.  u is the command sender. */
+
+static void privlist_add(User *u, int listid, const char *nick)
+{
+    NickInfo **list = (listid==LIST_ADMIN ? services_admins : services_opers);
+    int max = (listid==LIST_ADMIN ? MAX_SERVADMINS : MAX_SERVOPERS);
+    int *msgs = privlist_msgs[listid];
+    NickInfo *ni;
+    int i;
+
+    if (!(ni = findnick(nick))) {
+	notice_lang(s_OperServ, u, NICK_X_NOT_REGISTERED, nick);
+	return;
+    }
+    for (i = 0; i < max; i++) {
+	if (!list[i] || list[i] == ni)
+	    break;
+    }
+    if (list[i] == ni) {
+	notice_lang(s_OperServ, u, msgs[MSG_EXISTS], ni->nick);
+    } else if (i < max) {
+	list[i] = ni;
+	notice_lang(s_OperServ, u, msgs[MSG_ADDED], ni->nick);
+	if (readonly)
+	    notice_lang(s_OperServ, u, READ_ONLY_MODE);
+    } else {
+	notice_lang(s_OperServ, u, msgs[MSG_TOOMANY], max);
+    }
+}
+
+/*************************************************************************/
+
+/* Remove a nick from the Services admin/oper list. */
+
+static void privlist_rem(User *u, int listid, const char *nick)
+{
+    NickInfo **list = (listid==LIST_ADMIN ? services_admins : services_opers);
+    int max = (listid==LIST_ADMIN ? MAX_SERVADMINS : MAX_SERVOPERS);
+    int *msgs = privlist_msgs[listid];
+    NickInfo *ni;
+    int i;
+
+    if (!(ni = findnick(nick))) {
+	notice_lang(s_OperServ, u, NICK_X_NOT_REGISTERED, nick);
+	return;
+    }
+    for (i = 0; i < max; i++) {
+	if (list[i] == ni)
+	    break;
+    }
+    if (i < max) {
+	list[i] = NULL;
+	notice_lang(s_OperServ, u, msgs[MSG_REMOVED], ni->nick);
+	if (readonly)
+	    notice_lang(s_OperServ, u, READ_ONLY_MODE);
+    } else {
+	notice_lang(s_OperServ, u, msgs[MSG_NOTFOUND], ni->nick);
+    }
+}
+
 /*************************************************************************/
 /*********************** OperServ command functions **********************/
 /*************************************************************************/
@@ -526,7 +659,7 @@ static void do_global(User *u)
 	syntax_error(s_OperServ, u, "GLOBAL", OPER_GLOBAL_SYNTAX);
 	return;
     }
-#if HAVE_ALLWILD_NOTICE
+#ifdef HAVE_ALLWILD_NOTICE
     notice(s_GlobalNoticer, "$*", "%s", msg);
 #else
 # ifdef NETWORK_DOMAIN
@@ -578,6 +711,10 @@ static void do_stats(User *u)
 	    else
 		notice_lang(s_OperServ, u, OPER_STATS_AKILL_EXPIRE_NONE);
 	    return;
+	} else if (stricmp(extra, "RESET") == 0) {
+	    maxusercnt = usercnt;
+	    maxusertime = time(NULL);
+	    notice_lang(s_OperServ, u, OPER_STATS_RESET_USER_COUNT);
 	} else {
 	    notice_lang(s_OperServ, u, OPER_STATS_UNKNOWN_OPTION,
 			strupper(extra));
@@ -588,67 +725,22 @@ static void do_stats(User *u)
     tm = localtime(&maxusertime);
     strftime_lang(timebuf, sizeof(timebuf), u, STRFTIME_DATE_TIME_FORMAT, tm);
     notice_lang(s_OperServ, u, OPER_STATS_MAX_USERS, maxusercnt, timebuf);
-    if (days > 1) {
+    if (days >= 1) {
+	char *str = getstring(u->ni, days!=1 ? STR_DAYS : STR_DAY);
 	notice_lang(s_OperServ, u, OPER_STATS_UPTIME_DHMS,
-		days, hours, mins, secs);
-    } else if (days == 1) {
-	notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1DHMS,
-		days, hours, mins, secs);
+		days, str, hours, mins, secs);
     } else {
-	if (hours > 1) {
-	    if (mins != 1) {
-		if (secs != 1) {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_HMS,
-				hours, mins, secs);
-		} else {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_HM1S,
-				hours, mins, secs);
-		}
-	    } else {
-		if (secs != 1) {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_H1MS,
-				hours, mins, secs);
-		} else {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_H1M1S,
-				hours, mins, secs);
-		}
-	    }
-	} else if (hours == 1) {
-	    if (mins != 1) {
-		if (secs != 1) {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1HMS,
-				hours, mins, secs);
-		} else {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1HM1S,
-				hours, mins, secs);
-		}
-	    } else {
-		if (secs != 1) {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1H1MS,
-				hours, mins, secs);
-		} else {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1H1M1S,
-				hours, mins, secs);
-		}
-	    }
+	char *str1, *str2;
+	if (hours >= 1) {
+	    str1 = getstring(u->ni, hours!=1 ? STR_HOURS : STR_HOUR);
+	    str2 = getstring(u->ni, mins!=1 ? STR_MINUTES : STR_MINUTE);
+	    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_HM_MS,
+			hours, str1, mins, str2);
 	} else {
-	    if (mins != 1) {
-		if (secs != 1) {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_MS,
-				mins, secs);
-		} else {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_M1S,
-				mins, secs);
-		}
-	    } else {
-		if (secs != 1) {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1MS,
-				mins, secs);
-		} else {
-		    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_1M1S,
-				mins, secs);
-		}
-	    }
+	    str1 = getstring(u->ni, mins!=1 ? STR_MINUTES : STR_MINUTE);
+	    str2 = getstring(u->ni, secs!=1 ? STR_SECONDS : STR_SECOND);
+	    notice_lang(s_OperServ, u, OPER_STATS_UPTIME_HM_MS,
+			mins, str1, secs, str2);
 	}
     }
 
@@ -657,12 +749,15 @@ static void do_stats(User *u)
 	int i;
 
 	notice_lang(s_OperServ, u, OPER_STATS_BYTES_READ, total_read / 1024);
-	notice_lang(s_OperServ, u, OPER_STATS_BYTES_WRITTEN, 
+	notice_lang(s_OperServ, u, OPER_STATS_BYTES_WRITTEN,
 			total_written / 1024);
 
 	get_user_stats(&count, &mem);
 	notice_lang(s_OperServ, u, OPER_STATS_USER_MEM,
 			count, (mem+512) / 1024);
+	get_server_stats(&count, &mem);
+	notice_lang(s_OperServ, u, OPER_STATS_SERVER_MEM,
+			count, (mem+512) / 1024);
 	get_channel_stats(&count, &mem);
 	notice_lang(s_OperServ, u, OPER_STATS_CHANNEL_MEM,
 			count, (mem+512) / 1024);
@@ -672,6 +767,14 @@ static void do_stats(User *u)
 	get_chanserv_stats(&count, &mem);
 	notice_lang(s_OperServ, u, OPER_STATS_CHANSERV_MEM,
 			count, (mem+512) / 1024);
+	get_memoserv_stats(&count, &mem);
+	notice_lang(s_OperServ, u, OPER_STATS_MEMOSERV_MEM,
+			count, (mem+512) / 1024);
+#ifdef STATISTICS
+	get_statserv_stats(&count, &mem);
+	notice_lang(s_OperServ, u, OPER_STATS_STATSERV_MEM,
+			count, (mem+512) / 1024);
+#endif
 	count = 0;
 	if (CheckClones) {
 	    mem = sizeof(struct clone) * CLONE_DETECT_SIZE * 2;
@@ -692,15 +795,19 @@ static void do_stats(User *u)
 	get_news_stats(&count2, &mem2);
 	count += count2;
 	mem += mem2;
+#ifndef STREAMLINED
 	get_exception_stats(&count2, &mem2);
 	count += count2;
 	mem += mem2;
+#endif
 	notice_lang(s_OperServ, u, OPER_STATS_OPERSERV_MEM,
 			count, (mem+512) / 1024);
 
+#ifndef STREAMLINED
 	get_session_stats(&count, &mem);
 	notice_lang(s_OperServ, u, OPER_STATS_SESSIONS_MEM,
 			count, (mem+512) / 1024);
+#endif
     }
 }
 
@@ -735,15 +842,11 @@ static void do_os_mode(User *u)
     if (!(c = findchan(chan))) {
 	notice_lang(s_OperServ, u, CHAN_X_NOT_IN_USE, chan);
     } else if (c->bouncy_modes) {
-#if defined(IRC_DALNET) || defined(IRC_UNDERNET)
-	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES_U_LINE);
-#else
-	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES);
-#endif
+	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES_MESSAGE);
 	return;
     } else {
 	send_cmd(s_OperServ, "MODE %s %s", chan, modes);
-	if (WallOSMode)
+	if (WallOSChannel)
 	    wallops(s_OperServ, "%s used MODE %s on %s", u->nick, modes, chan);
 	*s = ' ';
 	argc = split_buf(chan, &argv, 1);
@@ -758,99 +861,64 @@ static void do_os_mode(User *u)
 static void do_clearmodes(User *u)
 {
     char *s;
-    int i;
-    char *argv[3];
     char *chan = strtok(NULL, " ");
     Channel *c;
     int all = 0;
-    int count;		/* For saving ban info */
-    char **bans;	/* For saving ban info */
-    struct c_userlist *cu, *next;
 
     if (!chan) {
 	syntax_error(s_OperServ, u, "CLEARMODES", OPER_CLEARMODES_SYNTAX);
     } else if (!(c = findchan(chan))) {
 	notice_lang(s_OperServ, u, CHAN_X_NOT_IN_USE, chan);
     } else if (c->bouncy_modes) {
-#if defined(IRC_DALNET) || defined(IRC_UNDERNET)
-	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES_U_LINE);
-#else
-	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES);
-#endif
+	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES_MESSAGE);
 	return;
     } else {
 	s = strtok(NULL, " ");
 	if (s) {
-	    if (strcmp(s, "ALL") == 0) {
+	    if (stricmp(s, "ALL") == 0) {
 		all = 1;
 	    } else {
 		syntax_error(s_OperServ,u,"CLEARMODES",OPER_CLEARMODES_SYNTAX);
 		return;
 	    }
 	}
-	if (WallOSClearmodes)
+	if (WallOSChannel)
 	    wallops(s_OperServ, "%s used CLEARMODES%s on %s",
 			u->nick, all ? " ALL" : "", chan);
 	if (all) {
-	    /* Clear mode +o */
-	    for (cu = c->chanops; cu; cu = next) {
-		next = cu->next;
-		argv[0] = sstrdup(chan);
-		argv[1] = sstrdup("-o");
-		argv[2] = sstrdup(cu->user->nick);
-		send_cmd(MODE_SENDER(s_ChanServ), "MODE %s %s :%s",
-			argv[0], argv[1], argv[2]);
-		do_cmode(s_ChanServ, 3, argv);
-		free(argv[2]);
-		free(argv[1]);
-		free(argv[0]);
-	    }
-
-	    /* Clear mode +v */
-	    for (cu = c->voices; cu; cu = next) {
-		next = cu->next;
-		argv[0] = sstrdup(chan);
-		argv[1] = sstrdup("-v");
-		argv[2] = sstrdup(cu->user->nick);
-		send_cmd(MODE_SENDER(s_ChanServ), "MODE %s %s :%s",
-			argv[0], argv[1], argv[2]);
-		do_cmode(s_ChanServ, 3, argv);
-		free(argv[2]);
-		free(argv[1]);
-		free(argv[0]);
-	    }
+	    clear_channel(c, CLEAR_UMODES, (void *)MODE_ALL);
+	    clear_channel(c, CLEAR_MODES | CLEAR_BANS | CLEAR_EXCEPTS, NULL);
+	    notice_lang(s_OperServ, u, OPER_CLEARMODES_ALL_DONE, chan);
+	} else {
+	    clear_channel(c, CLEAR_MODES | CLEAR_BANS | CLEAR_EXCEPTS, NULL);
+	    notice_lang(s_OperServ, u, OPER_CLEARMODES_DONE, chan);
 	}
+    }
+}
 
-	/* Clear modes */
-	send_cmd(MODE_SENDER(s_OperServ), "MODE %s -iklmnpst :%s",
-		chan, c->key ? c->key : "");
-	argv[0] = sstrdup(chan);
-	argv[1] = sstrdup("-iklmnpst");
-	argv[2] = c->key ? c->key : sstrdup("");
-	do_cmode(s_OperServ, 2, argv);
-	free(argv[0]);
-	free(argv[1]);
-	free(argv[2]);
-	c->key = NULL;
-	c->limit = 0;
+/*************************************************************************/
 
-	/* Clear bans */
-	count = c->bancount;
-	bans = smalloc(sizeof(char *) * count);
-	for (i = 0; i < count; i++)
-	    bans[i] = sstrdup(c->bans[i]);
-	for (i = 0; i < count; i++) {
-	    argv[0] = sstrdup(chan);
-	    argv[1] = sstrdup("-b");
-	    argv[2] = bans[i];
-	    send_cmd(MODE_SENDER(s_OperServ), "MODE %s %s :%s",
-			argv[0], argv[1], argv[2]);
-	    do_cmode(s_OperServ, 3, argv);
-	    free(argv[2]);
-	    free(argv[1]);
-	    free(argv[0]);
-	}
-	free(bans);
+/* Remove all users from a channel. */
+
+static void do_clearchan(User *u)
+{
+    char *chan = strtok(NULL, " ");
+    Channel *c;
+    char buf[BUFSIZE];
+
+    if (!chan) {
+	syntax_error(s_OperServ, u, "CLEARCHAN", OPER_CLEARCHAN_SYNTAX);
+    } else if (!(c = findchan(chan))) {
+	notice_lang(s_OperServ, u, CHAN_X_NOT_IN_USE, chan);
+    } else if (c->bouncy_modes) {
+	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES_MESSAGE);
+	return;
+    } else {
+	if (WallOSChannel)
+	    wallops(s_OperServ, "%s used CLEARCHAN on %s", u->nick, chan);
+	snprintf(buf, sizeof(buf), "CLEARCHAN by %s", u->nick);
+	clear_channel(c, CLEAR_USERS, buf);
+	notice_lang(s_OperServ, u, OPER_CLEARCHAN_DONE, chan);
     }
 }
 
@@ -874,23 +942,16 @@ static void do_os_kick(User *u)
     if (!(c = findchan(chan))) {
 	notice_lang(s_OperServ, u, CHAN_X_NOT_IN_USE, chan);
     } else if (c->bouncy_modes) {
-#if defined(IRC_DALNET) || defined(IRC_UNDERNET)
-	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES_U_LINE);
-#else
-	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES);
-#endif
+	notice_lang(s_OperServ, u, OPER_BOUNCY_MODES_MESSAGE);
 	return;
     }
     send_cmd(s_OperServ, "KICK %s %s :%s (%s)", chan, nick, u->nick, s);
-    if (WallOSKick)
+    if (WallOSChannel)
 	wallops(s_OperServ, "%s used KICK on %s/%s", u->nick, nick, chan);
-    argv[0] = sstrdup(chan);
-    argv[1] = sstrdup(nick);
-    argv[2] = sstrdup(s);
+    argv[0] = chan;
+    argv[1] = nick;
+    argv[2] = s;
     do_kick(s_OperServ, 3, argv);
-    free(argv[2]);
-    free(argv[1]);
-    free(argv[0]);
 }
 
 /*************************************************************************/
@@ -900,7 +961,6 @@ static void do_os_kick(User *u)
 static void do_admin(User *u)
 {
     char *cmd, *nick;
-    NickInfo *ni;
     int i;
 
     if (skeleton) {
@@ -917,28 +977,10 @@ static void do_admin(User *u)
 	    return;
 	}
 	nick = strtok(NULL, " ");
-	if (nick) {
-	    if (!(ni = findnick(nick))) {
-		notice_lang(s_OperServ, u, NICK_X_NOT_REGISTERED, nick);
-		return;
-	    }
-	    for (i = 0; i < MAX_SERVADMINS; i++) {
-		if (!services_admins[i] || services_admins[i] == ni)
-		    break;
-	    }
-	    if (services_admins[i] == ni) {
-		notice_lang(s_OperServ, u, OPER_ADMIN_EXISTS, ni->nick);
-	    } else if (i < MAX_SERVADMINS) {
-		services_admins[i] = ni;
-		notice_lang(s_OperServ, u, OPER_ADMIN_ADDED, ni->nick);
-	    } else {
-		notice_lang(s_OperServ, u, OPER_ADMIN_TOO_MANY, MAX_SERVADMINS);
-	    }
-	    if (readonly)
-		notice_lang(s_OperServ, u, READ_ONLY_MODE);
-	} else {
+	if (nick)
+	    privlist_add(u, LIST_ADMIN, nick);
+	else
 	    syntax_error(s_OperServ, u, "ADMIN", OPER_ADMIN_ADD_SYNTAX);
-	}
 
     } else if (stricmp(cmd, "DEL") == 0) {
 	if (!is_services_root(u)) {
@@ -946,26 +988,10 @@ static void do_admin(User *u)
 	    return;
 	}
 	nick = strtok(NULL, " ");
-	if (nick) {
-	    if (!(ni = findnick(nick))) {
-		notice_lang(s_OperServ, u, NICK_X_NOT_REGISTERED, nick);
-		return;
-	    }
-	    for (i = 0; i < MAX_SERVADMINS; i++) {
-		if (services_admins[i] == ni)
-		    break;
-	    }
-	    if (i < MAX_SERVADMINS) {
-		services_admins[i] = NULL;
-		notice_lang(s_OperServ, u, OPER_ADMIN_REMOVED, ni->nick);
-		if (readonly)
-		    notice_lang(s_OperServ, u, READ_ONLY_MODE);
-	    } else {
-		notice_lang(s_OperServ, u, OPER_ADMIN_NOT_FOUND, ni->nick);
-	    }
-	} else {
+	if (nick)
+	    privlist_rem(u, LIST_ADMIN, nick);
+	else
 	    syntax_error(s_OperServ, u, "ADMIN", OPER_ADMIN_DEL_SYNTAX);
-	}
 
     } else if (stricmp(cmd, "LIST") == 0) {
 	notice_lang(s_OperServ, u, OPER_ADMIN_LIST_HEADER);
@@ -986,7 +1012,6 @@ static void do_admin(User *u)
 static void do_oper(User *u)
 {
     char *cmd, *nick;
-    NickInfo *ni;
     int i;
 
     if (skeleton) {
@@ -1003,28 +1028,10 @@ static void do_oper(User *u)
 	    return;
 	}
 	nick = strtok(NULL, " ");
-	if (nick) {
-	    if (!(ni = findnick(nick))) {
-		notice_lang(s_OperServ, u, NICK_X_NOT_REGISTERED, nick);
-		return;
-	    }
-	    for (i = 0; i < MAX_SERVOPERS; i++) {
-		if (!services_opers[i] || services_opers[i] == ni)
-		    break;
-	    }
-	    if (services_opers[i] == ni) {
-		notice_lang(s_OperServ, u, OPER_OPER_EXISTS, ni->nick);
-	    } else if (i < MAX_SERVOPERS) {
-		services_opers[i] = ni;
-		notice_lang(s_OperServ, u, OPER_OPER_ADDED, ni->nick);
-	    } else {
-		notice_lang(s_OperServ, u, OPER_OPER_TOO_MANY, MAX_SERVOPERS);
-	    }
-	    if (readonly)
-		notice_lang(s_OperServ, u, READ_ONLY_MODE);
-	} else {
+	if (nick)
+	    privlist_add(u, LIST_OPER, nick);
+	else
 	    syntax_error(s_OperServ, u, "OPER", OPER_OPER_ADD_SYNTAX);
-	}
 
     } else if (stricmp(cmd, "DEL") == 0) {
 	if (!is_services_admin(u)) {
@@ -1032,26 +1039,10 @@ static void do_oper(User *u)
 	    return;
 	}
 	nick = strtok(NULL, " ");
-	if (nick) {
-	    if (!(ni = findnick(nick))) {
-		notice_lang(s_OperServ, u, NICK_X_NOT_REGISTERED, nick);
-		return;
-	    }
-	    for (i = 0; i < MAX_SERVOPERS; i++) {
-		if (services_opers[i] == ni)
-		    break;
-	    }
-	    if (i < MAX_SERVOPERS) {
-		services_opers[i] = NULL;
-		notice_lang(s_OperServ, u, OPER_OPER_REMOVED, ni->nick);
-		if (readonly)
-		    notice_lang(s_OperServ, u, READ_ONLY_MODE);
-	    } else {
-		notice_lang(s_OperServ, u, OPER_OPER_NOT_FOUND, ni->nick);
-	    }
-	} else {
+	if (nick)
+	    privlist_rem(u, LIST_OPER, nick);
+	else
 	    syntax_error(s_OperServ, u, "OPER", OPER_OPER_DEL_SYNTAX);
-	}
 
     } else if (stricmp(cmd, "LIST") == 0) {
 	notice_lang(s_OperServ, u, OPER_OPER_LIST_HEADER);
@@ -1067,14 +1058,59 @@ static void do_oper(User *u)
 
 /*************************************************************************/
 
+/* Obtain Services root privileges.  We check permissions here instead of
+ * letting run_cmd() do it for us so we can send out warnings when a
+ * non-admin tries to use the command.
+ */
+
+static void do_su(User *u)
+{
+    char *password = strtok(NULL, "");
+    int res;
+
+    if (!is_services_admin(u)) {
+	wallops(s_OperServ, "NOTICE: %s!%s@%s attempted to use SU command "
+		"(not Services admin)", u->nick, u->username, u->host);
+	notice_lang(s_OperServ, u, PERMISSION_DENIED);
+	return;
+    }
+
+    if (!password) {
+	syntax_error(s_OperServ, u, "SU", OPER_SU_SYNTAX);
+    } else if (skeleton) {
+	notice_lang(s_OperServ, u, OPER_SKELETON_MODE);
+    } else if (no_supass) {
+	notice_lang(s_OperServ, u, OPER_SU_NO_PASSWORD);
+    } else if ((res = check_password(password, supass)) < 0) {
+	notice_lang(s_OperServ, u, OPER_SU_FAILED);
+    } else if (res == 0) {
+	log("%s: Failed SU by %s!%s@%s",
+	    s_OperServ, u->nick, u->username, u->host);
+	wallops(s_OperServ, "NOTICE: Failed SU by %s!%s@%s",
+		u->nick, u->username, u->host);
+	bad_password(s_OperServ, u, "Services root");
+    } else {
+	/* The user must be a Services admin to get here so we know they
+	 * have a valid NickInfo. */
+	u->real_ni->status |= NS_SERVROOT;
+	if (WallSU)
+	    wallops(s_OperServ,
+		    "%s!%s@%s obtained Services super-user privileges",
+		    u->nick, u->username, u->host);
+	notice_lang(s_OperServ, u, OPER_SU_SUCCEEDED);
+    }
+}
+
+/*************************************************************************/
+
 /* Set various Services runtime options. */
 
 static void do_set(User *u)
 {
     char *option = strtok(NULL, " ");
-    char *setting = strtok(NULL, " ");
+    char *setting = strtok(NULL, "");
 
-    if (!option || !setting) {
+    if (!option || (!setting && stricmp(option, "SUPASS") != 0)) {
 	syntax_error(s_OperServ, u, "SET", OPER_SET_SYNTAX);
 
     } else if (stricmp(option, "IGNORE") == 0) {
@@ -1121,6 +1157,33 @@ static void do_set(User *u)
 	    notice_lang(s_OperServ, u, OPER_SET_DEBUG_ERROR);
 	}
 
+    } else if (stricmp(option, "SUPASS") == 0) {
+	int len;
+	char newpass[PASSMAX];
+
+	if (!is_services_root(u)) {
+	    notice_lang(s_OperServ, u, PERMISSION_DENIED);
+	    return;
+	}
+	if (!setting) {
+	    no_supass = 1;
+	    notice_lang(s_OperServ, u, OPER_SET_SUPASS_NONE);
+	    return;
+	}
+	len = strlen(setting);
+	if (len >= PASSMAX) {
+	    memset(setting+PASSMAX-1, 0, len-(PASSMAX-1));
+	    len = PASSMAX-1;
+	    notice_lang(s_OperServ, u, PASSWORD_TRUNCATED, len);
+	}
+	if (encrypt(setting, len, newpass, PASSMAX) < 0) {
+	    notice_lang(s_OperServ, u, OPER_SET_SUPASS_FAILED);
+	} else {
+	    no_supass = 0;
+	    memcpy(supass, newpass, PASSMAX);
+	    notice_lang(s_OperServ, u, OPER_SET_SUPASS_OK);
+	}
+
     } else {
 	notice_lang(s_OperServ, u, OPER_SET_UNKNOWN_OPTION, option);
     }
@@ -1132,23 +1195,18 @@ static void do_jupe(User *u)
 {
     char *jserver = strtok(NULL, " ");
     char *reason = strtok(NULL, "");
-    char buf[NICKMAX+16];
+    char buf[BUFSIZE];
 
     if (!jserver) {
 	syntax_error(s_OperServ, u, "JUPE", OPER_JUPE_SYNTAX);
     } else {
 	wallops(s_OperServ, "\2Juping\2 %s by request of \2%s\2.",
 		jserver, u->nick);
-	if (!reason) {
+	if (reason)
+	    snprintf(buf, sizeof(buf), "%s (%s)", reason, u->nick);
+	else
 	    snprintf(buf, sizeof(buf), "Jupitered by %s", u->nick);
-	    reason = buf;
-	}
-#ifdef IRC_UNDERNET_NEW
-	send_cmd(NULL, "SERVER %s 1 %lu %lu P09 :%s",
-		jserver, time(NULL), time(NULL), reason);
-#else
-	send_cmd(NULL, "SERVER %s 2 :%s", jserver, reason);
-#endif
+	send_server_remote(jserver, buf);
     }
 }
 
@@ -1201,39 +1259,27 @@ static void do_shutdown(User *u)
 
 static void do_restart(User *u)
 {
-#ifdef SERVICES_BIN
     quitmsg = malloc(31 + strlen(u->nick));
     if (!quitmsg)
 	quitmsg = "RESTART command received, but out of memory!";
     else
 	sprintf(quitmsg, "RESTART command received from %s", u->nick);
     raise(SIGHUP);
-#else
-    notice_lang(s_OperServ, u, OPER_CANNOT_RESTART);
-#endif
 }
 
 /*************************************************************************/
 
-/* XXX - this function is broken!! */
-
 static void do_listignore(User *u)
 {
     int sent_header = 0;
     IgnoreData *id;
-    int i;
 
-    notice(s_OperServ, u->nick, "Command disabled - it's broken.");
-    return;
-    
-    for (i = 0; i < 256; i++) {
-	for (id = ignore[i]; id; id = id->next) {
-	    if (!sent_header) {
-		notice_lang(s_OperServ, u, OPER_IGNORE_LIST);
-		sent_header = 1;
-	    }
-	    notice(s_OperServ, u->nick, "%ld %s", id->time, id->who);
+    for (id = first_ignore(); id; id = next_ignore()) {
+	if (!sent_header) {
+	    notice_lang(s_OperServ, u, OPER_IGNORE_LIST);
+	    sent_header = 1;
 	}
+	notice(s_OperServ, u->nick, "%ld %s", id->time, id->who);
     }
     if (!sent_header)
 	notice_lang(s_OperServ, u, OPER_IGNORE_LIST_EMPTY);
@@ -1262,10 +1308,10 @@ static void do_matchwild(User *u)
  * to prevent abuse and mistakes from being made - which might cause *.com to
  * be killed. It also makes it very quick and simple to use - which is usually
  * what you want when someone starts loading numerous clones. In addition to
- * killing the clones, we add a temporary AKILL to prevent them from 
+ * killing the clones, we add a temporary AKILL to prevent them from
  * immediately reconnecting.
  * Syntax: KILLCLONES nick
- * -TheShadow (29 Mar 1999) 
+ * -TheShadow (29 Mar 1999)
  */
 
 static void do_killclones(User *u)
@@ -1289,13 +1335,13 @@ static void do_killclones(User *u)
 
 	akillmask = smalloc(strlen(cloneuser->host) + 3);
 	sprintf(akillmask, "*@%s", strlower(cloneuser->host));
-	
+
 	user = firstuser();
 	while (user) {
 	    if (match_usermask(clonemask, user) != 0) {
 		tempuser = nextuser();
 		count++;
-		snprintf(killreason, sizeof(killreason), 
+		snprintf(killreason, sizeof(killreason),
 					"Cloning [%d]", count);
 		kill_user(NULL, user->nick, killreason);
 		user = tempuser;
@@ -1304,7 +1350,7 @@ static void do_killclones(User *u)
 	    }
 	}
 
-	add_akill(akillmask, akillreason, u->nick, 
+	add_akill(akillmask, akillreason, u->nick,
 			time(NULL) + KillClonesAkillExpire);
 
 	wallops(s_OperServ, "\2%s\2 used KILLCLONES for \2%s\2 killing "
